/*******************************************************************
 *
 *	File:		Arlo.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.demon.co.uk)
 *				Based on code from Otfried Cheong
 *
 *	(c) Copyright 2001, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include <e32std.h>
#include <e32base.h>
#include <e32hal.h>
#include <e32rom.h>
#include <e32ver.h>
#include <e32svr.h>

#include "ekern.h"
#include "arlo.h"

class TCapsArLoV1
{
public:
	TVersion version;
};


// version information
const TInt KMajorVersionNumber=0;
const TInt KMinorVersionNumber=0;
const TInt KBuildVersionNumber=0;


class DArLoDriver : public DLogicalDevice
{
public:
	DArLoDriver();
	virtual TInt Install();
	virtual void GetCaps(TDes8 &aDes) const;
	virtual DLogicalChannel *CreateL();

		// for some reason the default Remove returns KErrNotSupported
		// if we return KErrNon in Remove the machine reboots :-(
//	virtual TInt Remove(void){return KErrGeneral;}
};


class DArLo : public DLogicalChannel
{
public:
	DArLo(DLogicalDevice *aDevice);
protected:
	virtual void DoCancel(TInt aReqNo);
	virtual void DoRequest(TInt aReqNo,TAny *a1,TAny *a2);
	virtual TInt DoControl(TInt aFunction,TAny *a1,TAny *a2);
	TInt OriginalBoot(TInt aFunction, TAny* a1, TAny* a2);
	TInt BootS7(TInt aFunction, TAny* a1, TAny* a2, TAny* aCleanAddress);
};



DArLo::DArLo(DLogicalDevice *aDevice)
  : DLogicalChannel(aDevice)
{
//  SetBehaviour(0); // indicate we do not support asynchronous requests
}

void DArLo::DoCancel(TInt /* aReqNo */)
{
	INT3;
}

void DArLo::DoRequest(TInt /*aReqNo*/,TAny* /*a1*/,TAny* /*a2*/)
{
	INT3;
}




void FiddleDomainAccess()
{
#ifdef __MARM__
	asm("mov r0, #0xffffffff    ");
	asm("mcr p15, 0, r0, c3, c0 ");
#endif
}

#define __NAKED__ __declspec( naked )

#ifdef __MARM__
__NAKED__ TUint GetProcessorId()
{
	asm("  mrc	p15, 0, r0, c0, c0");		// @ get Processor ID
//	asm("mrs r0, cpsr");
	asm("  mov  pc, lr");  // return
}
#else
TUint GetProcessorId()
{
	return 0xf;
}
#endif


__NAKED__ TInt DArLo::OriginalBoot(TInt aFunction, TAny* a1, TAny* a2)
{
#ifdef __MARM__
  // handle any synchronous request function
  // can trash r0-r3 and ip
  // on entry:
  //   r0=this
  //   r1=aFunction
  //   r2=a1 (PhysMapAdr)
  //   r3=a2 (BootAdr)
  // on exit:
  //   r0=result
  //                                  

  // turn off interrupts
  asm("  mrs r0, cpsr");
  asm("  bic r0, r0, #0xdf");
  asm("  orr r0, r0, #0xd3");  // disable FIQ, IRQ, set SVC mode
  asm("  msr cpsr,r0");

  // load r0 with physical map address
  asm("  mov r0, r2");

  // turn off MMU (write zero into MMU register)
  asm("  mov R1,#070");      // not quite zero---don't want 26 bit mode
  asm("  mcr p15,0,R1,c1,c1");

  // we have two instructions here before hell breaks lose!
  // jump to R3
  asm("  mov pc,R3");
  // will never get here
#endif
}

const TInt	KCleanVariable = 0;
#define ROM_LINEAR_BASE 0

#if 1
__NAKED__ TInt DArLo::BootS7(TInt aFunction, TAny* a1, TAny* a2, TAny* aCleanAddress)
{
#ifdef __MARM__
// on entry:
//   r0=this
//   r1=aFunction
//   r2=a1 (PhysMapAdr)
//   r3=a2 (BootAdr)
//   r4=aCleanAddress

	// turn off interrupts
	asm("mrs r0, cpsr");
	asm("orr r0, r0, #0x60");  // disable FIQ, IRQ, don't fiddle with mode
	asm("msr cpsr, r0");

	// load r0 with physical map address
	asm("mov r0, r2");

#if 0
	asm("mov	r2, #1");
	asm("mov	r5, r4");
	asm("ldr	r6, =0xf5000000");		// FLUSHBASE
	asm("ldr	r1, [r5]");
	asm("ands	r1, r1, #1");
	asm("eor	r1, r1, #1");
	asm("str	r1, [r5]");
	asm("addne	r6, r6, #32768");		// FLUSH_OFFSET
//	asm("flush_110_dcache	r5, r6 r1");
//	.macro flush_110_dcache	rd, ra, re
	asm("add	r1, r6, #16384");		// @ only necessary for 16k
	asm("1001:	ldr	r5, [r6], #32");	// DCACHELINESIZE
	asm("teq	r1, r6");
	asm("bne	1001b");
	asm("mov	r6, #0");
	asm("teq	r2, #0");
	asm("mcrne	p15, 0, r6, c7, c5, 0");//		@ invalidate I cache
	asm("mcr	p15, 0, r6, c7, c10, 4");//		@ drain WB
#endif

#if 1
	asm("mov r1, #0");
	asm("mcr p15, 0, r1, c7, c7, 0");		// invalidate I,D caches
	asm("mcr p15, 0, r1, c7, c10, 4");		// drain WB
	asm("mcr p15, 0, r1, c8, c7, 0");		// invalidate I & D TLBs

	asm("mrc p15, 0, r1, c1, c0, 0");		// read ctrl register
	asm("bic r1, r1, #0x0000f");			// ............wcam
	asm("bic r1, r1, #0x00300");			// ......RS........
	asm("bic r1, r1, #0x01000");			// ...I............
	asm("mcr p15, 0, r1, c1, c0, 0");		// write ctrl register
	asm("mov r1, r1");
	asm("mov r1, r1");
#endif

#if 0
	asm("mov r1, #0");
	asm("mcr p15, 0, r1, c8, c7, 0 ");	// Flush TLBs
	asm("mcr p15, 0, r1, c7, c7, 0 ");	// Flush caches
	// turn off MMU (write zero into MMU register)
	asm("mov r1, #0");
	asm("mcr p15, 0, r1, c1, c1");
#endif


#if 0
	asm("mcr p15, 0, r0, c7, c10, 4");	// Drain Write Buffer
	asm("mov r0, #0xffffffff	");
	asm("mcr p15, 0, r0, c3, c0 ");		// disable MMU access checking
	asm("mov r0, #0x0			");
	asm("mcr p15, 0, r0, c8, c7, 0 ");	// Flush both TLBs
	asm("mcr p15, 0, r0, c7, c7, 0 ");	// Flush both caches
	asm("mov r0, r0 ");					// these two instructions will be skipped
	asm("mov r0, r0 ");
	asm("mov r0, #0x0 ");

	asm("mcr p15, 0, r0, c1, c0, 0");		// turn off the MMU
	asm("mov r0, r0 ");
	asm("mov r0, r0 ");
	asm("mov r0, r0 ");
#endif

	asm("mov pc, r3");
	// will never get here
#endif
}
#endif

#if 0
__NAKED__ TInt DArLo::BootS7(TInt aFunction, TAny* a1, TAny* a2, TAny* aCleanAddress)
{
			//; DCache flush
			//; (See Section 6.2.3 Software Dcache flush in SA-1110 Developers
			//; Manual)
        asm("ldr r0, =#0x88c00000");	// A virtual address for 0xe0000000 that has B=1/C=1
        asm("add r1, r0, #8192");		// 8KB cache / 32bytes = 256 lines l10
        asm("ldr r2, [r0], #32");		// load r2 from r0 then r0 =r0 + 32
        asm("teq r0, r1");				// have we reached our limit yet?
        asm("bne             l10");		// no, continue until done.

			// Mini DCache flush
        asm("ldr r0, =#0x88d00000");	// A virtual address for 0xe0000000 that has B=1/C=0
        asm("add r1, r0, #512");		// 512byte cache / 32 bytes = 16 lines l20
        asm("ldr r2, [r0], #32");//
        asm("teq r0, r1");//
        asm("bne l20");//

        asm("mov r0, #0");//
        asm("mcr p15, 0, r0, c7, c6, 0");//   ; flush D cache

        asm("mcr p15, 0, r0, c7, c10, 4");//  ; drain write buffer

//;       str             r0, [r4]                                ; interrupts off
        asm("mcr p15, 0, r0, c7, c7, 0");//   ; flush ID cache 
        

        //; At this point we must have everything in registers!
        asm("mcr p15, 0, r0, c1, c0, 0");//   ; disable MMU 
        asm("mov r0, r0");//
        asm("mov r0, r0");//
        asm("mov r0, r0");//
        asm("mov r0, r0");//
        asm("mov pc, r4");//
}
#endif

TInt DArLo::DoControl(TInt aFunction, TAny* a1, TAny* a2)
{
	INT3;

	switch (aFunction)
	{
		case KArloLddBootS5:
		case KArloLddBootS5mx:
			OriginalBoot(aFunction, a1, a2);
			break;

		case KArloLddBootS7:
			FiddleDomainAccess();
			BootS7(aFunction, a1, a2, &KCleanVariable);
			break;

		case KArloLddExamineArm4:
			FiddleDomainAccess();			// fall through
		case KArloLddExamineS5:
		{
			TUint*	val = (TUint*) a1;
			(*val) = *((TUint32*) *val);
			break;
		}

		case KArloLddProcId:
		{
			TUint*		p = (TUint*) a1;
			*p = GetProcessorId();
			break;
		}

		default:
			break;

	}
	return KErrNone;
}




DArLoDriver::DArLoDriver()
{
}

TInt DArLoDriver::Install()
{
	INT3;
	TPtrC name=KArloDriverName;       // our name, used by User::FreeLogicalDevice
	return SetName(&name);       // set our name and return error code
}

void DArLoDriver::GetCaps(TDes8 &aDes) const
{
	TVersion  version(KMajorVersionNumber, KMinorVersionNumber, KBuildVersionNumber);

	INT3;
	// standard capability-descriptor pattern
	TCapsArLoV1 b; // a structure
	b.version=version; // fill it with information
	aDes.FillZ(aDes.MaxLength()); // zero-pad the caller's descriptor
	aDes.Copy((TUint8 *)&b,Min(aDes.MaxLength(),sizeof(b)));
}

DLogicalChannel* DArLoDriver::CreateL()
{
	INT3;
	return new(ELeave) DArLo(this);
}


EXPORT_C DLogicalDevice *CreateLogicalDevice()
{
	INT3;
	DLogicalDevice *arlo = new(ELeave) DArLoDriver;
	return arlo;
}


// DLL entry point, required by all DLLs,
// called on thread or process attach/detach
GLDEF_C TInt E32Dll(TDllReason /*aReason*/)
{
  return KErrNone; // do nothing, return no error
}

